@@ -6,9 +6,10 @@ use ObjStore ':ADV';
 #require ObjStore::AV::Set; #?
 use base 'ObjStore::HV';
 use vars qw($VERSION);
-$VERSION = '1.02';
+$VERSION = '1.05';
 
 sub new {
+    use attrs 'method';
     my ($class, $where) = @_;
     croak "$class\->new(where)" if @_ != 2;
     my $o = $class->SUPER::new($where);
@@ -16,6 +17,7 @@ sub new {
 }
 
 sub add_index {
+    use attrs 'method';
     my ($o, $name, $index) = @_;
     croak "keys starting with underscore are reserved"
 	if $name =~ m/^_/;
@@ -40,37 +42,60 @@ sub add_index {
 }
 
 sub remove_index {
+    use attrs 'method';
     my ($o, $name) = @_;
     die "$o->remove_index($name): is not an index"
 	if !exists $o->{ $name };
     delete $o->{ $name };
     @{$$o{_allindices}} = grep($_ ne $name, @{$$o{_allindices}});
-    $$o{_primary} ||= $$o{_allindices}->[0]
-	if @{$$o{_allindices}};
-}
-
-sub index { $_[0]->{$_[1]}; }
-
-sub fetch { 
-    my $t=shift;
-    my $iname = shift;
-    my $i = $t->{ $iname };
-    croak "Can't find index '$iname'" if !$i;
-    my $c = $i->new_cursor;
-    if ($c->seek(@_)) {
-	$c->at;
+    if (@{$$o{_allindices}}) {
+	$$o{_primary} = $o->index($$o{_allindices}->[0]);
     } else {
-	();
+	$$o{_primary} = undef;
     }
 }
 
-sub primary {
-    my ($o) = @_;
-    return if !$$o{_primary};
-    $o->index($$o{_primary});
+sub index {
+    use attrs 'method';
+    $_[0]->{$_[1]};
+}
+
+sub fetch {
+    use attrs 'method';
+    my $t=shift;
+    my $iname = shift;
+    my $i = $t->{ $iname };
+    croak "can't find index '$iname'" if !$i;
+    my $c = $i->new_cursor;
+    if (!wantarray) {
+	return $c->seek(@_)? $c->at : undef;
+    } else {
+	my $pe = ObjStore::PathExam->new;
+	$pe->load_args(@_);
+	$c->step(1) if !$c->seek($pe); # exact match not needed
+	my @got;
+	while (my $e = $c->at) {
+	    last if $pe->compare($e) != 0;
+	    push @got, $e;
+	    $c->step(1);
+	}
+	@got;
+    }
+}
+
+sub at {
+    my ($o, $iname, $where) = @_;
+    my $x = $o->{$iname};
+    croak "Can't find index '$iname'" if !$x;
+    my $len = @$x;
+    return if $len == 0;
+    my $c = $x->new_cursor;
+    $c->moveto($where eq 'last'? $len-1 : $where);
+    $c->at();
 }
 
 sub anyx {
+    use attrs 'method';
     my ($o) = @_;
     if ($$o{_primary}) {
 	return $$o{_primary};
@@ -91,12 +116,14 @@ sub anyx {
 }
 
 sub rows {
+    use attrs 'method';
     my ($t) = @_;
     my $i = $t->anyx;
     $i? $i->count : 0;
 }
 
 sub map {
+    use attrs 'method';
     my ($t, $sub) = @_;
     my $x = $t->anyx;
     return if !$x;
@@ -104,6 +131,7 @@ sub map {
 }
 
 sub map_indices {
+    use attrs 'method';
     my ($o, $c) = @_;
     for my $i (@{$$o{_allindices}}) {
 	$c->( $$o{$i} );
@@ -111,11 +139,13 @@ sub map_indices {
 }
 
 sub add {
+    use attrs 'method';
     my ($t, $o) = @_;
     $t->map_indices(sub { shift->add($o) });
     defined wantarray ? $o : ();
 }
 sub remove {
+    use attrs 'method';
     my ($t, $o) = @_;
     $t->map_indices(sub { shift->remove($o) });
 }
@@ -126,6 +156,13 @@ sub compress {
 
 sub table { $_[0]; }
 
+sub primary {
+    carp "primary is depreciated";
+    my ($o) = @_;
+    return if !$$o{_primary};
+    $o->index($$o{_primary});
+}
+
 package ObjStore::Table3::Database;
 use Carp;
 use ObjStore;
@@ -160,7 +197,8 @@ __END__
 
 =head1 SYNOPSIS
 
-  cd table-test ObjStore::Table3::Database
+  my $table = ObjStore::Table3->new($near);
+  $table->add_index('name', sub { ObjStore::Index->new($table, path => 'name') }};
 
 =head1 DESCRIPTION
 
@@ -172,19 +210,10 @@ A table is no more than a collection of indices (as opposed to a some
 sort of heavy-weight object).  Think of it like an event manager for
 indices.
 
-Be aware that index cursors may only be used by one thread at a time.
-Therefore, it is not particularly useful to store pre-created cursors
-in a database.  It's probably faster just to create them transiently
-when needed.
-
 =head2 API
 
 =over 4
 
-=item * $t->primary()
-
-Returns the primary index.
-
 =item * $t->anyx
 
 Returns a non-empty index.
@@ -204,7 +233,11 @@ Returns the index named $index_name.
 =item * $t->fetch($index_name, @keys)
 
 Returns the record resulting from looking up @keys in the index named
-$index_name.
+$index_name.  Also works in an array context.
+
+=item * $t->at($index_name, $offset)
+
+The $offset should either be numeric or 'last'.
 
 =item * $t->add_index($name, $index)
 
@@ -219,35 +252,16 @@ Calls $coderef->($index) on each index.
 
 =back
 
-=head2 Representation Independent API
+=head2 CAVEAT
 
-A database can be seen as table, and/or tables can be stored within a
-database.  The implementation is only slightly different in either
-case.  To smooth things over, an accessor method is provided that
-always returns the top-level hash of the table.
-
-=over 4
-
-=item * $t->table
-
-Returns the top-level hash.
-
-=back
-
-=head1 MIGRATION
-
-Expand migration options?
+Be aware that index cursors may only be used by one process/thread at
+a time.  Therefore, it is usually not helpful to store pre-created
+cursors in a database.
 
 =head1 TODO
 
-=over 4
-
-=item * INTERFACE
-
-The interface will evolve as perl supports more overload-type
-features.
-
-=back
+I'm fairly satisfied at this point.  Some thing may still be improved.
+Ideas welcome!
 
 =head1 AUTHOR
 
